@@ -1,7 +1,7 @@
 // Copyright 2015 Ted Mielczarek. See the COPYRIGHT
 // file at the top-level directory of this distribution.
 
-use failure::{format_err, Error};
+use failure::format_err;
 use log::warn;
 use nom::IResult::*;
 use nom::*;
@@ -18,9 +18,10 @@ use std::str::FromStr;
 use minidump_common::traits::IntoRangeMapSafe;
 
 use crate::sym_file::types::*;
+use crate::SymbolError;
 
 enum Line<'a> {
-    Info,
+    Info(Info),
     File(u32, &'a str),
     Public(PublicSymbol),
     Function(Function),
@@ -61,7 +62,19 @@ named!(module_line<&[u8], ()>,
     || {}
 ));
 
-// Matches an INFO record.
+// Matches an INFO LINE record.
+named!(
+    info_url<&[u8], Info>,
+    chain!(
+        tag!("INFO URL") ~
+        space ~
+        url: map_res!(not_line_ending, str::from_utf8) ~
+        my_eol,
+          ||{ Info::Url(url.to_string()) }
+    )
+);
+
+// Matches other INFO records.
 named!(
     info_line,
     chain!(
@@ -166,6 +179,8 @@ chain!(
 // Matches a STACK WIN record.
 named!(stack_win_line<&[u8], WinFrameType>,
   chain!(
+    // Use this for debugging the parser:
+    // line: peek!(map_res!(not_line_ending, str::from_utf8)) ~
     tag!("STACK WIN") ~
     space ~
     // Frame type, currently ignored.
@@ -192,7 +207,23 @@ named!(stack_win_line<&[u8], WinFrameType>,
     rest: map_res!(not_line_ending, str::from_utf8) ~
     my_eol ,
       || {
-          let program_string_or_base_pointer = if has_program_string {
+          // Sometimes has_program_string is just wrong. We could try to infer which one is right
+          // but this is rare enough that it's better to just play it safe and discard the input.
+          let really_has_program_string = ty == b"4";
+          if really_has_program_string != has_program_string {
+            let kind = match ty {
+              b"4" => "FrameData",
+              b"0" => "Fpo",
+              _ => "Unknown Type!",
+            };
+            warn!("STACK WIN entry had inconsistent values for type and has_program_string, discarding corrupt entry");
+            // warn!("  {}", &line);
+            warn!("  type: {} ({}), has_program_string: {}, final_arg: {}", str::from_utf8(ty).unwrap_or(""), kind, has_program_string, rest);
+
+            return WinFrameType::Unhandled;
+          }
+
+          let program_string_or_base_pointer = if really_has_program_string {
               WinStackThing::ProgramString(rest.to_string())
           } else {
               WinStackThing::AllocatesBasePointer(rest == "1")
@@ -272,12 +303,13 @@ named!(stack_cfi_lines<&[u8], StackInfoCfi>,
 // Parse any of the line data that can occur in the body of a symbol file.
 named!(line<&[u8], Line>,
   alt!(
-    info_line => { |_| Line::Info } |
+    info_url => { Line::Info } |
+    info_line => { |_| Line::Info(Info::Unknown) } |
     file_line => { |(i,f)| Line::File(i, f) } |
-    public_line => { |p| Line::Public(p) } |
-    func_lines => { |f| Line::Function(f) } |
-    stack_win_line => { |s| Line::StackWin(s) } |
-    stack_cfi_lines => { |s| Line::StackCfi(s) }
+    public_line => { Line::Public } |
+    func_lines => { Line::Function } |
+    stack_win_line => { Line::StackWin } |
+    stack_cfi_lines => { Line::StackCfi }
 ));
 
 // Return a `SymbolFile` given a vec of `Line` data.
@@ -288,9 +320,15 @@ fn symbol_file_from_lines(lines: Vec<Line<'_>>) -> SymbolFile {
     let mut stack_cfi = vec![];
     let mut stack_win_framedata: Vec<StackInfoWin> = vec![];
     let mut stack_win_fpo: Vec<StackInfoWin> = vec![];
+    let mut url = None;
     for line in lines {
         match line {
-            Line::Info => {}
+            Line::Info(Info::Url(cached_url)) => {
+                url = Some(cached_url);
+            }
+            Line::Info(Info::Unknown) => {
+                // Don't care
+            }
             Line::File(id, filename) => {
                 files.insert(id, filename.to_string());
             }
@@ -324,7 +362,9 @@ fn symbol_file_from_lines(lines: Vec<Line<'_>>) -> SymbolFile {
                                     // addr: 1, len: 2
                                     // addr: 4, len: 6
                                     last.size = (info.address - last.address) as u32;
-                                } else {
+                                } else if last.memory_range().unwrap() != memory_range {
+                                    // We silently drop identical ranges because sometimes
+                                    // duplicates happen, but we complain for non-trivial duplicates.
                                     warn!(
                                         "STACK WIN entry had bad intersections, dropping it {:?}",
                                         info
@@ -374,6 +414,16 @@ fn symbol_file_from_lines(lines: Vec<Line<'_>>) -> SymbolFile {
             .into_iter()
             .map(|s| (s.memory_range(), s))
             .into_rangemap_safe(),
+        // Will get filled in by the caller
+        url,
+        // TODO
+        ambiguities_repaired: 0,
+        // TODO
+        ambiguities_discarded: 0,
+        // TODO
+        corruptions_discarded: 0,
+        // TODO
+        cfi_eval_corruptions: 0,
     }
 }
 
@@ -386,7 +436,7 @@ named!(symbol_file<&[u8], SymbolFile>,
 );
 
 /// Parse a `SymbolFile` from `bytes`.
-pub fn parse_symbol_bytes(bytes: &[u8]) -> Result<SymbolFile, Error> {
+pub fn parse_symbol_bytes(bytes: &[u8]) -> Result<SymbolFile, SymbolError> {
     match symbol_file(bytes) {
         Done(rest, symfile) => {
             if rest == b"" {
@@ -396,7 +446,7 @@ pub fn parse_symbol_bytes(bytes: &[u8]) -> Result<SymbolFile, Error> {
                 let next_line = rest
                     .split(|b| *b == b'\r')
                     .next()
-                    .map(|bytes| String::from_utf8_lossy(bytes))
+                    .map(String::from_utf8_lossy)
                     .unwrap_or(Cow::Borrowed(""));
                 Err(format_err!(
                     "Failed to parse file, next line was: `{}`",
@@ -407,13 +457,16 @@ pub fn parse_symbol_bytes(bytes: &[u8]) -> Result<SymbolFile, Error> {
         Error(e) => Err(format_err!("Failed to parse file: {}", e)),
         Incomplete(_) => Err(format_err!("Failed to parse file: incomplete data")),
     }
+    .map_err(SymbolError::ParseError)
 }
 
 /// Parse a `SymbolFile` from `path`.
-pub fn parse_symbol_file(path: &Path) -> Result<SymbolFile, Error> {
-    let mut f = File::open(path)?;
+pub fn parse_symbol_file(path: &Path) -> Result<SymbolFile, SymbolError> {
+    let mut f = File::open(path)
+        .map_err(|e| SymbolError::LoadError(format_err!("Failed to open file: {}", e)))?;
     let mut bytes = vec![];
-    f.read_to_end(&mut bytes)?;
+    f.read_to_end(&mut bytes)
+        .map_err(|e| SymbolError::LoadError(format_err!("Failed to read file: {}", e)))?;
     parse_symbol_bytes(&bytes)
 }
 
@@ -456,6 +509,14 @@ fn test_info_line2() {
     assert_eq!(info_line(line), Done(rest, bits));
 }
 
+#[test]
+fn test_info_url() {
+    let line = b"INFO URL https://www.example.com\n";
+    let url = "https://www.example.com".to_string();
+    let rest = &b""[..];
+    assert_eq!(info_url(line), Done(rest, Info::Url(url)));
+}
+
 #[test]
 fn test_file_line() {
     let line = b"FILE 1 foo.c\n";
@@ -1031,3 +1092,73 @@ FILE"[..]
         "Should fail to parse empty file"
     );
 }
+
+#[test]
+fn test_parse_stack_win_inconsistent() {
+    // Various cases where the has_program_string value is inconsistent
+    // with the type of the STACK WIN entry.
+    //
+    // type=0 (FPO) should go with has_program_string==0 (false)
+    // type=4 (FrameData) should go with has_program_string==1 (true)
+    //
+    // Only 4d93e and 8d93e are totally valid.
+    //
+    // Current policy is to discard all the other ones, but all the cases
+    // are here in case we decide on a more sophisticated heuristic.
+
+    let bytes = b"MODULE Windows x86 D3096ED481217FD4C16B29CD9BC208BA0 firefox-bin
+FILE 0 foo.c
+STACK WIN 0 1d93e 4 4 0 0 10 0 0 1 1
+STACK WIN 0 2d93e 4 4 0 0 10 0 0 1 0
+STACK WIN 0 3d93e 4 4 0 0 10 0 0 1 prog string
+STACK WIN 0 4d93e 4 4 0 0 10 0 0 0 1
+STACK WIN 4 5d93e 4 4 0 0 10 0 0 0 1
+STACK WIN 4 6d93e 4 4 0 0 10 0 0 0 0
+STACK WIN 4 7d93e 4 4 0 0 10 0 0 0 prog string
+STACK WIN 4 8d93e 4 4 0 0 10 0 0 1 prog string
+";
+    let sym = parse_symbol_bytes(&bytes[..]).unwrap();
+
+    assert_eq!(sym.win_stack_framedata_info.ranges_values().count(), 1);
+    let ws = sym
+        .win_stack_framedata_info
+        .ranges_values()
+        .map(|&(_, ref s)| s)
+        .collect::<Vec<_>>();
+    {
+        let stack = &ws[0];
+        assert_eq!(stack.address, 0x8d93e);
+        assert_eq!(stack.size, 0x4);
+        assert_eq!(stack.prologue_size, 0x4);
+        assert_eq!(stack.epilogue_size, 0);
+        assert_eq!(stack.parameter_size, 0);
+        assert_eq!(stack.saved_register_size, 0x10);
+        assert_eq!(stack.local_size, 0);
+        assert_eq!(stack.max_stack_size, 0);
+        assert_eq!(
+            stack.program_string_or_base_pointer,
+            WinStackThing::ProgramString("prog string".to_string())
+        );
+    }
+    assert_eq!(sym.win_stack_fpo_info.ranges_values().count(), 1);
+    let ws = sym
+        .win_stack_fpo_info
+        .ranges_values()
+        .map(|&(_, ref s)| s)
+        .collect::<Vec<_>>();
+    {
+        let stack = &ws[0];
+        assert_eq!(stack.address, 0x4d93e);
+        assert_eq!(stack.size, 0x4);
+        assert_eq!(stack.prologue_size, 0x4);
+        assert_eq!(stack.epilogue_size, 0);
+        assert_eq!(stack.parameter_size, 0);
+        assert_eq!(stack.saved_register_size, 0x10);
+        assert_eq!(stack.local_size, 0);
+        assert_eq!(stack.max_stack_size, 0);
+        assert_eq!(
+            stack.program_string_or_base_pointer,
+            WinStackThing::AllocatesBasePointer(true)
+        );
+    }
+}
